		.586
		.MODEL flat, C
		.MMX

		OPTION PROLOGUE:NONE
		OPTION EPILOGUE:NONE

		.CODE

; Map table of supported conversions.
; Address this in C as a [2][6][6] array:
;	First index is 1 if MMX is OK, else 0
;	Second index is source (nChannels - 1)
;	Third index is destination (nChannels - 1)
; NULL (0) means the conversion isn't supported

PUBLIC	C MapTab

MapTab:
	; Non-MMX conversions
		; src is 1 channel
			DWORD	OFFSET asm_convert_1_to_1
			DWORD	OFFSET asm_convert_2_to_2	; liba52 upmixes
			DWORD	0, 0, 0, 0
		; src is 2 channels
			DWORD	OFFSET asm_convert_1_to_1	; liba52 downmixes
			DWORD	OFFSET asm_convert_2_to_2
			DWORD	0, 0, 0, 0
		; src is 3 channels
			DWORD	OFFSET asm_convert_1_to_1
			DWORD	OFFSET asm_convert_2_to_2
			DWORD	OFFSET asm_convert_3_to_3
			DWORD	0, 0, 0
		; src is 4 channels
			DWORD	OFFSET asm_convert_1_to_1
			DWORD	OFFSET asm_convert_2_to_2
			DWORD	0, OFFSET asm_convert_4_to_4
			DWORD	0, 0
		; src is 5 channels
			DWORD	OFFSET asm_convert_1_to_1
			DWORD	OFFSET asm_convert_2_to_2
			DWORD	0, 0, OFFSET asm_convert_5_to_5
			DWORD	0
		; src is 6 channels
			DWORD	OFFSET asm_convert_1_to_1
			DWORD	OFFSET asm_convert_2_to_2
			DWORD	0, 0, 0, OFFSET asm_convert_6_to_6

	; MMX conversions
		; src is 1 channel
			DWORD	OFFSET mmx_convert_1_to_1
			DWORD	OFFSET mmx_convert_2_to_2
			DWORD	0, 0, 0, 0
		; src is 2 channels
			DWORD	OFFSET mmx_convert_1_to_1
			DWORD	OFFSET mmx_convert_2_to_2
			DWORD	0, 0, 0, 0
		; src is 3 channels
			DWORD	OFFSET mmx_convert_1_to_1
			DWORD	OFFSET mmx_convert_2_to_2
			DWORD	OFFSET mmx_convert_3_to_3
			DWORD	0, 0, 0
		; src is 4 channels
			DWORD	OFFSET mmx_convert_1_to_1
			DWORD	OFFSET mmx_convert_2_to_2
			DWORD	0, OFFSET mmx_convert_4_to_4
			DWORD	0, 0
		; src is 5 channels
			DWORD	OFFSET mmx_convert_1_to_1
			DWORD	OFFSET mmx_convert_2_to_2
			DWORD	0, 0, OFFSET mmx_convert_5_to_5
			DWORD	0
		; src is 6 channels
			DWORD	OFFSET mmx_convert_1_to_1
			DWORD	OFFSET mmx_convert_2_to_2
			DWORD	0, 0, 0, OFFSET mmx_convert_6_to_6

mmxconst	LABEL QWORD
			DWORD	43C00000h, 43C00000h

; These must be the same as in a52.h
	A52_CHANNEL			= 0
	A52_MONO			= 1
	A52_STEREO			= 2
	A52_3F				= 3
	A52_2F1R			= 4
	A52_3F1R			= 5
	A52_2F2R			= 6
	A52_3F2R			= 7
	A52_CHANNEL1		= 8
	A52_CHANNEL2		= 9
	A52_DOLBY			= 10
	A52_CHANNEL_MASK	= 15
	A52_LFE				= 16

; This is how liba52 creates its "flags" value:
;	static uint8_t lfeon[8] = {16, 16, 4, 4, 4, 1, 4, 1};
;	int acmod = buf[6] >> 5;
;	*flags = (
;		(((buf[6] & 0xf8) == 0x50)? A52_DOLBY: acmod)
;		|
;		((buf[6] & lfeon[acmod])? A52_LFE: 0)
;	);
;
; From this, only certain combinations are possible:
;	acmod ranges from 0 to 7
;	if ((buf[6] & 0xF8) == 0x50) {	; 0101 0...
;		acmod must be 2, therefore lfeon[acmod] is 4
;		the result is A52_DOLBY, or (A52_DOLBY | A52_LFE)
;	} else {
;	the result is acmod, or (acmod | A52_LFE)
;	}
;
; This is from the Dolby spec:
;	if ((acmod & 1) && (acmod != 1)) /* if 3 front channels */ {cmixlev} 2
;	if (acmod & 4)	/* if a surround channel exists */ {surmixlev} 2
;	if (acmod == 2) /* if in 2/0 mode */ {dsurmod} 2
;
; Ultimately these are 9 mutually exclusive possibilities:
;	000l ....	1+1		A52_CHANNEL
;	001l ....	1/0		A52_MONO
;	0100 xl..	2/0		A52_STEREO
;	0101 xl..	2/0		A52_STEREO | A52_DOLBY
;	100s sl..	2/1		A52_2F1R
;	101c cssl	3/1		A52_3F1R
;	110s sl..	2/2		A52_2F2R
;	111c cssl	3/2		A52_3F2R
; Any of these may be combined with the A52_LFE flag.
; A surround channel exists only if (flags & 4)
;
; Note that A52_DOLBY includes the A52_STEREO flag.

clip1 	MACRO	reg1:REQ
	; Valid range is 43BF8000h to 43C07FFFh
	; (this clipper definitely needs improvement!)

		.IF (reg1 > 43C07FFFh)
			mov		reg1, 00007FFFh
		.ELSEIF (reg1 < 43BF8000h)
			mov		reg1, 00008000h
		.ENDIF
ENDM

clip2	MACRO	reg1:REQ, reg2:REQ
	; clip two registers back-to-back
	clip1	reg1
	clip1	reg2
ENDM

IsMMX	PROC

		pushfd
		mov		eax, 200000h
		xor		eax, [esp]	; toggle ID flag
		push	eax
		popfd
		pushfd
		pop		eax
		xor		eax, [esp]
		popfd
		and		eax, 200000h
		.IF !ZERO?			; CPUID is supported
			push	ebx
			push	ecx
			push	edx
			xor		eax, eax
			cpuid
			test	eax, eax
			.IF !ZERO?		; EAX must be at least 1!
				mov		eax, 1
				cpuid
				xor		eax, eax
				shr		edx, 24		; CF = IA MMX bit
				adc		eax, eax
			.ENDIF
			pop		edx
			pop		ecx
			pop		ebx
		.ENDIF
		ret

IsMMX	ENDP

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Destination and Source are 1 channel

asm_convert_1_to_1  PROC PRIVATE
		push	esi
		push	edi
		mov		esi, [esp+12]	; src
		mov		edi, [esp+16]	; dst
		xor		ecx, ecx

		align	16
		.REPEAT
			mov		eax, [esi]		; C0
			mov		edx, [esi+4]	; C1
			clip2	eax, edx
			shl		eax, 16
			add		esi, 8
			shld	edx, eax, 16	; [C1][C0]
			mov		[edi], edx
			add		edi, 4
			sub		cl, 2		; did 2 samples
		.UNTIL ZERO?

		pop		edi
		pop		esi
		ret
asm_convert_1_to_1	ENDP

mmx_convert_1_to_1  PROC PRIVATE
	; convert floating mono to 16-bit mono
		mov		ecx, [esp+4]
		mov		edx, [esp+8]
		movq	mm7, mmxconst
		xor		eax, eax

		align	16
		.REPEAT
			movq		mm0, [ecx]		; [  C1  ][  C0  ]

			movq		mm1, [ecx+8]	; [  C3  ][  C2  ]
			psubd		mm0, mm7

			movq		mm2, [ecx+16]	; [  C5  ][  C4  ]
			psubd		mm1, mm7

			movq		mm3, [ecx+24]	; [  C7  ][  C6  ]
			packssdw	mm0, mm1		; [C3][C2][C1][C0]

			psubd		mm2, mm7
			psubd		mm3, mm7

			movq		[edx], mm0
			packssdw	mm2, mm3		; [C7][C6][C5][C4]

			add			ecx, 32
			sub			al, 8			; did 8 samples

			movq		[edx+8], mm2

			lea			edx, [edx+16]
		.UNTIL ZERO?

		emms
		ret
mmx_convert_1_to_1	ENDP

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Source and Destination are 2 channels
; Source is one of the following:
;	A52_MONO + A52_LFE			return FC-LF
;	A52_STEREO or A52_DOLBY		return FL-FR

asm_convert_2_to_2  PROC PRIVATE
		push	esi
		push	edi
		mov		esi, [esp+12]	; src
		push	ebx
		mov		edi, [esp+20]	; dst
		mov		ebx, 1024
		xor		ecx, ecx

		test	dword ptr [esp+24], A52_LFE
		jz		loop1
		add		esi, ebx
		neg		ebx

        align	16
loop1:      mov		eax, [esi]		; L
			mov		edx, [ebx+esi]	; R
			clip2	eax, edx
			shl		eax, 16
			add		esi, 4
			shld	edx, eax, 16
			mov		[edi], edx
			add		edi, 4
			dec		cl 		; did 1 sample
			jnz		loop1

        pop		ebx
		pop		edi
		pop		esi
		ret
asm_convert_2_to_2	ENDP

mmx_convert_2_to_2  PROC PRIVATE
		push	ebx
		mov		ecx, [esp+8]
		mov		ebx, [esp+12]
		lea		edx, [ecx+1024]
		movq	mm7, mmxconst
		xor		eax, eax

		test	dword ptr [esp+16], A52_LFE
		jz		loop1

	; liba52 would return LF-FC, so swap channels
		xchg	ecx, edx

        align	16
loop1:		movq		mm0, [ecx]			; [  L1  ][  L0  ]

			movq		mm1, [ecx+8]		; [  L3  ][  L2  ]
			psubd		mm0, mm7

			movq		mm2, [edx]			; [  R1  ][  R0  ]
			psubd		mm1, mm7

			movq		mm4, mm0			; [  L1  ][  L0  ]
			psubd		mm2, mm7

			movq		mm3, [edx+8]		; [  R3  ][  R2  ]
			punpckldq	mm0, mm2			; [  R0  ][  L0  ]

			punpckhdq	mm4, mm2			; [  R1  ][  L1  ]
			psubd		mm3, mm7

			movq		mm2, mm1			; [  L3  ][  L2  ]
			packssdw	mm0, mm4			; [R1][L1][R0][L0]

			punpckldq	mm1, mm3			; [  R2  ][  L2  ]
			add			ecx, 16

			movq		[ebx], mm0
			punpckhdq	mm2, mm3			; [  R3  ][  L3  ]

			packssdw	mm1, mm2			; [R3][L3][R2][L2]
			add			edx, 16

			sub			al, 4				; did 4 samples

			movq		[ebx+8], mm1

			lea			ebx, [ebx+16]
			jnz			loop1

        pop		ebx
		emms
		ret

mmx_convert_2_to_2	ENDP

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Source and Destination are 3 channels
; Source is one of the following:
;	A52_STEREO + A52_LFE	LF,FL,FR: return 1,2,0
;	A52_3F					FL,FC,FR: return 0,2,1
;	A52_2F1R				FL,FR,BC: return 0,1,2

CONVERT3	MACRO	CA, CB, CC
	align	16
	.REPEAT		; [CC1 CB1] [CA1 CC0] [CB0 CA0]
		mov		eax, [esi+CA+0]
		mov		edx, [esi+CB+0]
		clip2	eax, edx
		shl		eax, 16				; [CA0][...]
		mov		ebx, [esi+CC+0]
		shld	edx, eax, 16		; [CB0][CA0]
		mov		eax, [esi+CA+4]
		mov		[edi], edx
		add		edi, 4
		clip2	ebx, eax
		shl		ebx, 16				; [CC0][...]
		mov		edx, [esi+CB+4]
		shld	eax, ebx, 16		; [CA1][CC0]
		mov		ebx, [esi+CC+4]
		mov		[edi], eax
		add		edi, 4
		clip2	edx, ebx
		shl		edx, 16				; [CB1][...]
		add		esi, 8
		shld	ebx, edx, 16		; [CC1][CB1]
		mov		[edi], ebx
		add		edi, 4
		sub		cl, 2	; did 2 samples
	.UNTIL ZERO?
ENDM

asm_convert_3_to_3  PROC PRIVATE
		push	esi
		push	edi
		mov		esi, [esp+12]	; src
		push	ebx
		mov		edi, [esp+20]	; dst
		xor		ecx, ecx

		mov		eax, [esp+24]
		.IF (eax & A52_LFE)
			CONVERT3	1024, 2048, 0
		.ELSEIF (eax == A52_3F)
			CONVERT3	0, 2048, 1024
		.ELSE	; A52_2F1R
			CONVERT3	0, 1024, 2048
		.ENDIF

		pop		ebx
		pop		edi
		pop		esi
		ret
asm_convert_3_to_3	ENDP

MMXCONVERT3	MACRO	CA, CB, CC
	xor		eax, eax
	align	16
	.REPEAT
		movq		mm0, [ecx+CA]		; [ CA1 ][ CA0 ]
		movq		mm1, [ecx+CB]		; [ CB1 ][ CB0 ]
		psubd		mm0, mm7
		movq		mm2, [ecx+CC]		; [ CC1 ][ CC0 ]
		psubd		mm1, mm7
		psubd		mm2, mm7
	; we want     [CA1][CC0][CB0][CA0]
		movq		mm3, mm2       		; [ CC1 ][ CC0 ]
		psllq		mm2, 32				; [ CC0 ][ ... ]
		punpckhdq	mm2, mm0			; [ CA1 ][ CC0 ]
		punpckldq	mm0, mm1			; [ CB0 ][ CA0 ]
		packssdw	mm0, mm2			; [CA1][CC0][CB0][CA0]
		movq		mm2, [ecx+CA+8]		; [ CA3 ][ CA2 ]
		punpckhdq	mm1, mm3			; [ CC1 ][ CB1 ]
	; now we want [CB2][CA2][CC1][CB1]
		movq		mm3, [ecx+CB+8]		; [ CB3 ][ CB2 ]
		psubd		mm2, mm7
		movq		mm4, [ecx+CC+8]		; [ CC3 ][ CC2 ]
		psubd		mm3, mm7
		movq		mm5, mm2			; [ CA3 ][ CA2 ]
		psubd		mm4, mm7
		punpckldq	mm2, mm3			; [ CB2 ][ CA2 ]
	; last we want [CC3][CB3][CA3][CC2]
		movq		[edx], mm0
		packssdw	mm1, mm2			; [CB2][CA2][CC1][CB1]
		punpckhdq	mm3, mm4			; [ CC3 ][ CB3 ]
		psllq		mm4, 32				; [ CC2 ][ ... ]
		punpckhdq	mm4, mm5			; [ CA3 ][ CC2 ]
		movq		[edx+8], mm1
		packssdw	mm4, mm3			; [CC3][CB3][CA3][CC2]
		add			ecx, 16
		sub			al, 4		; did 4 samples
		movq		[edx+16], mm4
		lea			edx, [edx+24]
	.UNTIL ZERO?
ENDM

mmx_convert_3_to_3  PROC PRIVATE
		mov		ecx, [esp+4]	; src
		mov		edx, [esp+8]	; dst
		movq	mm7, mmxconst

		mov		eax, [esp+12]
		.IF (eax & A52_LFE)
			MMXCONVERT3	1024, 2048, 0
		.ELSEIF (eax == A52_3F)
			MMXCONVERT3	0, 2048, 1024
		.ELSE	; A52_2F1R
			MMXCONVERT3	0, 1024, 2048
		.ENDIF
		
		emms
		ret
mmx_convert_3_to_3	ENDP

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Source and Destination are 4 channels
; Source is one of the following:
;	A52_2F1R + A52_LFE		LF,FL,FR,BC: return 1,2,0,3
;	A52_3F + A52_LFE		LF,FL,FC,FR: return 1,3,2,0
;	A52_2F2R				FL,FR,BL,BR: return 0,1,2,3
;	A52_3F1R				FL,FC,FR,BC: return 0,2,1,3

CONVERT4	MACRO	CA, CB, CC, CD
	align	16
	.REPEAT
		mov		eax, [esi+CA]	; [  CA  ]
		mov		edx, [esi+CB]	; [  CB  ]
		clip2	eax, edx
		shl		eax, 16			; [CA][..]
		mov		ebx, [esi+CC]	; [  CC  ]
		shld	edx, eax, 16	; [CB][CA]
		mov		eax, [esi+CD]	; [  CD  ]
		mov		[edi], edx
		add		edi, 4
		clip2	ebx, eax
		shl		ebx, 16			; [CC][..]
		add		esi, 4
		shld	eax, ebx, 16	; [CD][CC]
		mov		[edi], eax
		add		edi, 4
		dec		cl		; did 1 sample
	.UNTIL ZERO?
ENDM

asm_convert_4_to_4  PROC PRIVATE
	; convert 4-channel floating to 16-bit
		push	esi
		push	edi
		mov		esi, [esp+12]	; src
		push	ebx
		mov		edi, [esp+20]	; dst
		xor		ecx, ecx

		mov		eax, [esp+24]
		.IF (eax == (A52_2F1R or A52_LFE))
			CONVERT4	1024, 2048, 0, 3072
		.ELSEIF (eax == (A52_3F or A52_LFE))
			CONVERT4	1024, 3072, 2048, 0
		.ELSEIF (eax == A52_2F2R)
			CONVERT4	0, 1024, 2048, 3072
		.ELSE	; A52_3F1R
			CONVERT4	0, 2048, 1024, 3072
		.ENDIF

		pop		ebx
		pop		edi
		pop		esi
		ret
asm_convert_4_to_4	ENDP

MMXCONVERT4	MACRO	CA, CB, CC, CD
	xor		eax, eax
	align	16
	.REPEAT
		movq		mm0, [ecx+CA]		; [ CA1 ][ CA0 ]
		movq		mm1, [ecx+CB]		; [ CB1 ][ CB0 ]
		psubd		mm0, mm7
		movq		mm2, [ecx+CC]		; [ CC1 ][ CC0 ]
		psubd		mm1, mm7
		movq		mm3, [ecx+CD]		; [ CD1 ][ CD0 ]
		psubd		mm2, mm7
		movq		mm4, mm0			; [ CA1 ][ CA0 ]
		psubd		mm3, mm7
	; we want     [CD0][CC0][CB0][CA0]
		punpckldq	mm0, mm1			; [ CB0 ][ CA0 ]
		movq		mm5, mm2			; [ CC1 ][ CC0 ]
		punpckldq	mm2, mm3			; [ CD0 ][ CC0 ]
		packssdw	mm0, mm2			; [CD0][CC0][CB0][CA0]
		add			ecx, 8
	; now we want [CD1][CC1][CB1][CA1]
		punpckhdq	mm4, mm1            ; [ CB1 ][ CA1 ]
		punpckhdq	mm5, mm3			; [ CD1 ][ CC1 ]
		packssdw	mm4, mm5			; [CD1][CC1][CB1][CA1]
		sub			al, 2		; did 2 samples
		movq		[edx], mm0
		movq		[edx+8], mm4
		lea			edx, [edx+16]
	.UNTIL ZERO?
ENDM

mmx_convert_4_to_4  PROC PRIVATE
	; convert 4-channel floating to 16-bit
		mov		ecx, [esp+4]
		mov		edx, [esp+8]
		movq	mm7, mmxconst

		mov		eax, [esp+12]
		.IF (eax == (A52_2F1R or A52_LFE))
			MMXCONVERT4	1024, 2048, 0, 3072
		.ELSEIF (eax == (A52_3F or A52_LFE))
			MMXCONVERT4	1024, 3072, 2048, 0
		.ELSEIF (eax == A52_2F2R)
			MMXCONVERT4	0, 1024, 2048, 3072
		.ELSE	; A52_3F1R
			MMXCONVERT4	0, 2048, 1024, 3072
		.ENDIF

		emms
		ret
mmx_convert_4_to_4	ENDP

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Source and Destination are 5 channels
; Source is one of the following:
;	A52_2F2R + A52_LFE		LF,FL,FR,BL,BR: return 1,2,3,4,0
;	A52_3F1R + A52_LFE		LF,FL,FC,FR,BC: return 1,3,2,0,4
;	A52_3F2R				FL,FC,FR,BL,BR: return 0,2,1,3,4

CONVERT5	MACRO	CA, CB, CC, CD, CE
	align	16
	.REPEAT
		mov		eax, [esi+CA+0]	; [   CA0  ]
		mov		edx, [esi+CB+0]	; [   CB0  ]
		clip2	eax, edx
		shl		eax, 16			; [CA0][...]
		mov		ebx, [esi+CC+0]	; [   CC0  ]
		shld	edx, eax, 16	; [CB0][CA0]
		mov		eax, [esi+CD+0]	; [   CD0  ]
		mov		[edi], edx
		add		edi, 4
		clip2	ebx, eax
		shl		ebx, 16			; [CC0][...]
		mov		edx, [esi+CE+0]	; [   CE0  ]
		shld	eax, ebx, 16	; [CD0][CC0]
		mov		ebx, [esi+CA+4]	; [   CA1  ]
		mov		[edi], eax
		add		edi, 4
		clip2	edx, ebx
		shl		edx, 16			; [CE0][...]
		mov		eax, [esi+CB+4]	; [   CB1  ]
		shld	ebx, edx, 16	; [CA1][CE0]
		mov		edx, [esi+CC+4]	; [   CC1  ]
		mov		[edi], ebx
		add		edi, 4
		clip2	eax, edx
		shl		eax, 16			; [CB1][...]
		mov		ebx, [esi+CD+4]	; [   CD1  ]
		shld	edx, eax, 16	; [CC1][CB1]
		mov		eax, [esi+CE+4]	; [   CE1  ]
		mov		[edi], edx
		add		edi, 4
		clip2	ebx, eax
		shl		ebx, 16			; [CD1][...]
		add		esi, 8
		shld	eax, ebx, 16	; [CE1][CD1]
		mov		[edi], eax
		add		edi, 4
		sub		cl, 2		; did 2 samples
	.UNTIL ZERO?
ENDM

asm_convert_5_to_5  PROC PRIVATE
	; convert 5-channel floating to 16-bit
		push	esi
		push	edi
		mov		esi, [esp+12]	; src
		push	ebx
		mov		edi, [esp+20]	; dst
		xor		ecx, ecx

		mov		eax, [esp+24]
		.IF (eax == (A52_2F2R or A52_LFE))
			CONVERT5	1024, 2048, 3072, 4096, 0
		.ELSEIF (eax == (A52_3F1R or A52_LFE))
			CONVERT5	1024, 3072, 2048, 0, 4096
		.ELSE	; A52_3F2R
			CONVERT5	0, 2048, 1024, 3072, 4096
		.ENDIF

		pop		ebx
		pop		edi
		pop		esi
		ret
asm_convert_5_to_5	ENDP

MMXCONVERT5	MACRO	CA, CB, CC, CD, CE
	xor		eax, eax
	align	16
	.REPEAT
		movq		mm0, [ecx+CA+0]		; [ CA1 ][ CA0 ]
		movq		mm1, [ecx+CB+0]		; [ CB1 ][ CB0 ]
		psubd		mm0, mm7
		movq		mm2, [ecx+CC+0]		; [ CC1 ][ CC0 ]
		psubd		mm1, mm7
		movq		mm3, [ecx+CD+0]		; [ CD1 ][ CD0 ]
		psubd		mm2, mm7
		movq		mm4, [ecx+CE+0]		; [ CE1 ][ CE0 ]
		psubd		mm3, mm7

	; we want     [CD0][CC0][CB0][CA0]
		movq		mm5, mm0			; [ CA1 ][ CA0 ]
		psubd		mm4, mm7
		movq		mm6, mm2			; [ CC1 ][ CC0 ]
		psrlq		mm0, 32				; [ ... ][ CA1 ]
		punpckldq	mm5, mm1			; [ CB0 ][ CA0 ]
		punpckldq	mm6, mm3			; [ CD0 ][ CC0 ]
		packssdw	mm5, mm6			; [CD0][CC0][CB0][CA0]
		movq		[edx], mm5

	; now we want [CC1][CB1][CA1][CE0]
		movq		mm5, mm4			; [ CE1 ][ CE0 ]
		movq		mm6, mm1			; [ CB1 ][ CB0 ]
		punpckldq	mm5, mm0			; [ CA1 ][ CE0 ]
		punpckhdq	mm6, mm2			; [ CC1 ][ CB1 ]
		packssdw	mm5, mm6			; [CC1][CB1][CA1][CE0]
		movq		[edx+8], mm5

	; now we want [CB2][CA2][CE1][CD1]
		movq		mm0, [ecx+CA+8]		; [ CA3 ][ CA2 ]
		movq		mm5, mm3			; [ CD1 ][ CD0 ]
		movq		mm1, [ecx+CB+8]		; [ CB3 ][ CB2 ]
		psubd		mm0, mm7
		movq		mm2, [ecx+CC+8]		; [ CC3 ][ CC2 ]
		psubd		mm1, mm7
		movq		mm3, [ecx+CD+8]		; [ CD3 ][ CD2 ]
		movq		mm6, mm0			; [ CA3 ][ CA2 ]
		punpckhdq	mm5, mm4			; [ CE1 ][ CD1 ]
		punpckldq	mm6, mm1			; [ CB2 ][ CA2 ]
		movq		mm4, [ecx+CE+8]		; [ CE3 ][ CE2 ]
		psubd		mm2, mm7
		packssdw	mm5, mm6			; [CB2][CA2][CE1][CD1]
		psubd		mm3, mm7
		movq		[edx+16], mm5

	; now we want [CA3][CE2][CD2][CC2]
		movq		mm5, mm2			; [ CC3 ][ CC2 ]
		psrlq		mm0, 32				; [ ... ][ CA3 ]
		movq		mm6, mm4			; [ CE3 ][ CE2 ]
		punpckldq	mm5, mm3			; [ CD2 ][ CC2 ]
		punpckldq	mm6, mm0			; [ CA3 ][ CE2 ]
		packssdw	mm5, mm6			; [CA3][CE2][CD2][CC2]
		movq		[edx+24], mm5

	; last we want [CE3][CD3][CC3][CB3]
		movq		mm5, mm1			; [ CB3 ][ CB2 ]
		movq		mm6, mm3			; [ CD3 ][ CD2 ]
		punpckhdq	mm5, mm2			; [ CC3 ][ CB3 ]
		punpckhdq	mm6, mm4			; [ CE3 ][ CD3 ]
		packssdw	mm5, mm6			; [CE3][CD3][CC3][CB3]
		movq		[edx+32], mm5

		add			ecx, 16
		add			edx, 40
		sub			al, 4		; did 4 samples
	.UNTIL ZERO?
ENDM

mmx_convert_5_to_5  PROC PRIVATE
	; convert 5-channel floating to 16-bit
		mov		ecx, [esp+4]
		mov		edx, [esp+8]
		movq	mm7, mmxconst

		mov		eax, [esp+12]
		.IF (eax == (A52_2F2R or A52_LFE))
			MMXCONVERT5	1024, 2048, 3072, 4096, 0
		.ELSEIF (eax == (A52_3F1R or A52_LFE))
			MMXCONVERT5	1024, 3072, 2048, 0, 4096
		.ELSE	; A52_3F2R
			MMXCONVERT5	0, 2048, 1024, 3072, 4096
		.ENDIF

		emms
		ret
mmx_convert_5_to_5	ENDP


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Source and Destination are 6 channels
; Source is A52_3F2R + A52_LFE
; LF,FL,FC,FR,BL,BR: return 1,3,2,0,4,5

asm_convert_6_to_6  PROC PRIVATE
	; convert 6-channel floating to 16-bit
		push	esi
		push	edi
		mov		esi, [esp+12]	; src
		push	ebx
		mov		edi, [esp+20]	; dst
		xor		ecx, ecx

		align	16
		.REPEAT
			mov		eax, [esi+1024*1]	; [  FL  ]
			mov		edx, [esi+1024*3]	; [  FR	 ]
			clip2	eax, edx
			shl		eax, 16				; [FL][..]
			mov		ebx, [esi+1024*2]	; [  FC	 ]
			shld	edx, eax, 16		; [FR][FL]
			mov		eax, [esi+1024*0]	; [  LF  ]
			mov		[edi], edx
			add		edi, 4
			clip2	ebx, eax
			shl		ebx, 16				; [FC][..]
			mov		edx, [esi+1024*4]	; [  BL  ]
			shld	eax, ebx, 16		; [LF][FC]
			mov		ebx, [esi+1024*5]	; [  BR  ]
			mov		[edi], eax
			add		edi, 4
			clip2	edx, ebx
			shl		edx, 16				; [BL][..]
			add		esi, 4
			shld	ebx, edx, 16		; [BR][BL]
			mov		[edi], ebx
			add		edi, 4
			dec		cl		; did 1 sample
		.UNTIL ZERO?

		pop		ebx
		pop		edi
		pop		esi
		ret
asm_convert_6_to_6	ENDP

mmx_convert_6_to_6  PROC PRIVATE
	; convert 6-channel floating to 16-bit
		mov		ecx, [esp+4]
		mov		edx, [esp+8]
		movq	mm7, mmxconst
		xor		eax, eax

		align	16
		.REPEAT
			movq		mm0, [ecx+1024*1]	; [ FL1 ][ FL0 ]
			movq		mm1, [ecx+1024*2]	; [ FC1 ][ FC0 ]
			psubd		mm0, mm7
			movq		mm2, [ecx+1024*3]	; [ FR1 ][ FR0 ]
			psubd		mm1, mm7
			movq		mm3, [ecx+1024*0]	; [ LF1 ][ LF0 ]
			psubd		mm2, mm7
			psubd		mm3, mm7
		; we want [LF0][FC0][FR0][FL0]
			movq		mm4, mm0			; [ FL1 ][ FL0 ]
			punpckldq	mm0, mm2			; [ FR0 ][ FL0 ]
			movq		mm5, mm1			; [ FC1 ][ FC0 ]
			punpckldq	mm1, mm3			; [ LF0 ][ FC0 ]
			movq		mm6, [ecx+1024*4]	; [ BL1 ][ BL0 ]
			packssdw	mm0, mm1			; [LF0][FC0][FR0][FL0]
		; we want [FR1][FL1][BR0][BL0]
			movq		mm1, [ecx+1024*5]	; [ BR1 ][ BR0 ]
			psubd		mm6, mm7
			psubd		mm1, mm7
			punpckhdq	mm4, mm2			; [ FR1 ][ FL1 ]
			movq		mm2, mm6			; [ BL1 ][ BL0 ]
			punpckldq	mm6, mm1			; [ BR0 ][ BL0 ]
			packssdw	mm6, mm4			; [FR1][FL1][BR0][BL0]
			movq		[edx], mm0
		; we want [BR1][BL1][LF1][FC1]
			punpckhdq	mm2, mm1			; [ BR1 ][ BL1 ]
			punpckhdq	mm5, mm3			; [ LF1 ][ FC1 ]
			packssdw	mm5, mm2			; [BR1][BL1][LF1][FC1]
			add			ecx, 8
			sub			al, 2		; did 2 samples
			movq		[edx+8], mm6
			movq		[edx+16], mm5
			lea			edx, [edx+24]
		.UNTIL ZERO?

		emms
		ret
mmx_convert_6_to_6	ENDP

		END
